# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2019 ZhongdaoWang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Darknet-53 backbone for JDE model.

Modifications include:
- Remove training related code such as:
    - classifier
    - loss_names
    - losses
    - test_emb
    - uniform initialization of batch norm
- Refactor to remove unused code
    - enumerate in forward()
    - "maxpool" in _create_nodes
- Refactor for proper type hinting
    - renamed one of layer_i to layer_indices in forward()
- Refactor in _create_nodes to reduce the number of local variables
- Use the nn.Upsample instead of the custom one since it no longer gives
    deprecated warning
- Use nn.Identity instead of custom EmptyLayer
    - relevant PR: https://github.com/pytorch/pytorch/pull/19249
- Removed yolo_layer_count since layer member variable has been removed in
    YOLOLayer as it's not used
"""

from typing import Any, Dict, List, Tuple

import torch
import torch.nn as nn

from peekingduck.pipeline.nodes.model.jdev1.jde_files.network_blocks import YOLOLayer


class Darknet(nn.Module):
    """YOLOv3 object detection model.

    Args:
        cfg_dict (List[Dict[str, Any]]): Model architecture
            configurations.
        device (torch.device): The device which a `torch.Tensor` is on or
            will be allocated.
        num_identities (int): Number of identities, e.g., number of unique
            pedestrians. Uses 14455 for JDE according to the original code.
    """

    def __init__(
        self, cfg_dict: List[Dict[str, Any]], device: torch.device, num_identities: int
    ) -> None:
        super().__init__()
        self.module_defs = cfg_dict
        self.device = device
        self.module_defs[0]["nID"] = num_identities
        self.img_size = (
            int(self.module_defs[0]["width"]),
            int(self.module_defs[0]["height"]),
        )
        self.emb_dim = int(self.module_defs[0]["embedding_dim"])
        self.hyperparams, self.module_list = _create_modules(
            self.module_defs, self.device
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=invalid-name
        """Defines the computation performed at every call.

        Args:
            inputs (torch.Tensor): Input from the previous layer.

        Returns:
            (torch.Tensor): A dictionary of tensors with keys corresponding to
                `self.out_features`.
        """
        layer_outputs: List[torch.Tensor] = []
        outputs = []

        for module_def, module in zip(self.module_defs, self.module_list):
            module_type = module_def["type"]
            if module_type in ["convolutional", "upsample", "maxpool"]:
                x = module(x)
            elif module_type == "route":
                layer_indices = list(map(int, module_def["layers"].split(",")))
                if len(layer_indices) == 1:
                    x = layer_outputs[layer_indices[0]]
                else:
                    x = torch.cat([layer_outputs[i] for i in layer_indices], 1)
            elif module_type == "shortcut":
                x = layer_outputs[-1] + layer_outputs[int(module_def["from"])]
            elif module_type == "yolo":
                x = module[0](x, self.img_size)
                outputs.append(x)
            layer_outputs.append(x)

        return torch.cat(outputs, 1)


def _create_modules(
    module_defs: List[Dict[str, Any]], device: torch.device
) -> Tuple[Dict[str, Any], nn.ModuleList]:
    """Constructs module list of layer blocks from module configuration in
    `module_defs`

    NOTE: Each `module_def` in `module_defs` is parsed as a dictionary
    containing string values. As a result, "1" can sometimes represent True
    instead of the number of the key. We try to do == "1" instead of implicit
    boolean by converting it to int.

    Args:
        module_defs (List[Dict[str, Any]]): A list of module definitions.
        device (torch.device): The device which a `torch.Tensor` is on or
                will be allocated.

    Returns:
        (Tuple[Dict[str, Any], nn.ModuleList]): A tuple containing a dictionary
            of model hyperparameters and a ModuleList containing the modules
            in the model.
    """
    hyperparams = module_defs.pop(0)
    output_filters = [int(hyperparams["channels"])]
    module_list = nn.ModuleList()
    for i, module_def in enumerate(module_defs):
        module_type = module_def["type"]
        modules = nn.Sequential()
        if module_type == "convolutional":
            has_batch_norm = module_def["batch_normalize"] == "1"
            filters = int(module_def["filters"])
            kernel_size = int(module_def["size"])
            modules.add_module(
                f"conv_{i}",
                nn.Conv2d(
                    in_channels=output_filters[-1],
                    out_channels=filters,
                    kernel_size=kernel_size,
                    stride=int(module_def["stride"]),
                    padding=(kernel_size - 1) // 2 if module_def["pad"] == "1" else 0,
                    bias=not has_batch_norm,
                ),
            )
            if has_batch_norm:
                modules.add_module(f"batch_norm_{i}", nn.BatchNorm2d(filters))
            if module_def["activation"] == "leaky":
                modules.add_module(f"leaky_{i}", nn.LeakyReLU(0.1))
        elif module_type == "upsample":
            modules.add_module(
                f"upsample_{i}",
                nn.Upsample(scale_factor=int(module_def["stride"]), mode="nearest"),
            )
        elif module_type == "route":
            filters = sum(
                [
                    output_filters[i + 1 if i > 0 else i]
                    for i in map(int, module_def["layers"].split(","))
                ]
            )
            modules.add_module(f"route_{i}", nn.Identity())
        elif module_type == "shortcut":
            filters = output_filters[int(module_def["from"])]
            modules.add_module(f"shortcut_{i}", nn.Identity())
        elif module_type == "yolo":
            # Extract anchors
            anchor_dims = iter(map(float, module_def["anchors"].split(",")))
            # This lets us do pairwise() with no overlaps
            anchors = list(zip(anchor_dims, anchor_dims))
            anchors = [anchors[i] for i in map(int, module_def["mask"].split(","))]
            # Define detection layer
            modules.add_module(
                f"yolo_{i}",
                YOLOLayer(
                    anchors,
                    int(module_def["classes"]),
                    int(hyperparams["nID"]),
                    int(hyperparams["embedding_dim"]),
                    device,
                ),
            )

        # Register module list and number of output filters
        module_list.append(modules)
        output_filters.append(filters)

    return hyperparams, module_list



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Preprocessing functions for HRNet
"""

from typing import Tuple

import cv2
import numpy as np


def tlwh2xywh(inputs: np.ndarray, aspect_ratio: float) -> np.ndarray:
    """Converts from [t, l, w, h] to [x, y, w1, h] format.

    (t, l) is the coordinates of the top left corner, w is the width, and h is
    the height. (x, y) is the coordinates of the center. The bounding box is
    adjusted to meet the input aspect ratio, w1 is the new width and h1 is the
    new height.

    [x, y, w1, h1] is calculated as:
    x = t + w * 0.5
    y = l + h * 0.5
    w1 = {
        h * aspect_ratio, if w < h * aspect_ratio
        w               , otherwise
    }
    h1 = {
        w / aspect_ratio, if w > h * aspect_ratio
        h               , otherwise
    }

    Args:
        inputs (np.ndarray): Input bounding boxes (2-d array) each with the
            format `(top left x, top left y, width, height)`.
        bboxes (np.ndarray): Array of bboxes in the form of top left
            (x, y, w, h).
        aspect_ratio (float): w:h ratio for the new box.

    Returns:
        (np.ndarray): Array of bboxes in center (x, y, w1, h1) format that
        meets the aspect ratio.
    """
    outputs = np.empty_like(inputs)
    outputs[:, 0] = inputs[:, 0] + inputs[:, 2] * 0.5
    outputs[:, 1] = inputs[:, 1] + inputs[:, 3] * 0.5
    outputs[:, 2] = inputs[:, 2]
    outputs[:, 3] = inputs[:, 3]

    target_width = inputs[:, 3] * aspect_ratio

    tall_bboxes = inputs[:, 2] < target_width
    outputs[tall_bboxes, 2] = inputs[tall_bboxes, 3] * aspect_ratio

    wide_bboxes = inputs[:, 2] > target_width
    outputs[wide_bboxes, 3] = inputs[wide_bboxes, 2] / aspect_ratio

    return outputs


def crop_and_resize(
    frame: np.ndarray, bboxes: np.ndarray, out_size: Tuple[int, int]
) -> Tuple[np.ndarray, np.ndarray]:
    """Crop a region from frame specified by its center and size. The
    cropped region is resized to out_size.

    Args:
        frame (np.ndarray): Image in numpy array.
        bboxes (np.ndarray): Bboxes center (x, y, w, h) coordinates.
        out_size (tuple): Cropped region will be resized to out_size.

    Returns:
        (Tuple[np.ndarray, np.ndarray]): The resized and cropped region array
        and the affine transform matrix to map a point in cropped image
        coordinate space to source frame coordinate space.
    """
    translate_x = bboxes[:, 0] - (bboxes[:, 2] - 1) * 0.5
    translate_y = bboxes[:, 1] - (bboxes[:, 3] - 1) * 0.5
    scale_x = bboxes[:, 2] / out_size[0]
    scale_y = bboxes[:, 3] / out_size[1]
    zero_mat = np.zeros((len(translate_x),))

    x_mat = np.column_stack((scale_x, zero_mat, translate_x))
    y_mat = np.column_stack((zero_mat, scale_y, translate_y))
    affine_matrices = np.concatenate((x_mat, y_mat), axis=1)
    affine_matrices = affine_matrices.reshape((-1, 2, 3))

    transformed_images = [
        cv2.warpAffine(
            frame, x, out_size, flags=cv2.INTER_LINEAR | cv2.WARP_INVERSE_MAP
        )
        for x in affine_matrices
    ]
    return transformed_images, affine_matrices



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of bboxes for efficientdet
"""

import logging
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.model_process import (
    postprocess_boxes,
    preprocess_image,
)
from peekingduck.utils.graph_functions import load_graph


class Detector:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Detector class to handle detection of bboxes for efficientdet"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: Dict[int, str],
        detect_ids: List[int],
        model_type: int,
        num_classes: int,
        model_file: Dict[int, str],
        model_nodes: Dict[str, List[str]],
        image_size: Dict[int, int],
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.class_names = class_names
        self.model_type = model_type
        self.num_classes = num_classes
        self.model_path = model_dir / model_file[self.model_type]
        self.model_nodes = model_nodes
        self.image_size = image_size[self.model_type]
        self.score_threshold = score_threshold

        self.detect_ids = detect_ids
        self.efficient_det = self._create_efficient_det_model()

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Efficientdet bbox prediction function

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            boxes (np.ndarray): array of detected bboxes
            labels (np.ndarray): array of labels
            scores (np.ndarray): array of scores
        """
        img_shape = image.shape[:2]
        image, scale = self._preprocess(image)

        # run network
        graph_input = tf.convert_to_tensor(
            np.expand_dims(image, axis=0), dtype=tf.float32
        )
        boxes, scores, labels = self.efficient_det(x=graph_input)
        network_output = (
            np.squeeze(boxes.numpy()),
            np.squeeze(scores.numpy()),
            np.squeeze(labels.numpy()),
        )

        boxes, labels, scores = self._postprocess(network_output, scale, img_shape)

        return boxes, labels, scores

    def _create_efficient_det_model(self) -> tf.keras.Model:
        model = load_graph(
            str(self.model_path),
            inputs=self.model_nodes["inputs"],
            outputs=self.model_nodes["outputs"],
        )
        self.logger.info(
            "EfficientDet model loaded with following configs:\n\t"
            f"Model type: D{self.model_type}\n\t"
            f"IDs being detected: {self.detect_ids}\n\t"
            f"Score threshold: {self.score_threshold}"
        )

        return model

    def _postprocess(
        self,
        network_output: Tuple[np.ndarray, np.ndarray, np.ndarray],
        scale: float,
        img_shape: Tuple[int, int],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Postprocessing of detected bboxes for efficientdet

        Args:
            network_output (list): list of boxes, scores and labels from network
            scale (float): scale the image was resized to
            img_shape (Tuple[int, int]): height of original image

        Returns:
            boxes (np.ndarray): postprocessed array of detected bboxes
            scores (np.ndarray): postprocessed array of scores
            labels (np.ndarray): postprocessed array of labels
        """
        img_h, img_w = img_shape
        boxes, scores, labels = network_output
        boxes = postprocess_boxes(boxes, scale, img_h, img_w)

        # Filter by confidence score
        score_filter = np.where(scores[:] > self.score_threshold)[0]
        boxes = boxes[score_filter]
        labels = labels[score_filter]
        scores = scores[score_filter]

        # Filter by detect ID
        detect_filter = np.where(np.isin(labels, self.detect_ids))
        boxes = boxes[detect_filter]
        labels = labels[detect_filter]
        scores = scores[detect_filter]

        if labels.size:
            labels = np.vectorize(self.class_names.get)(labels)
        return boxes, labels, scores

    def _preprocess(self, image: np.ndarray) -> Tuple[np.ndarray, float]:
        """Preprocessing function for efficientdet

        Args:
            image (np.ndarray): Image in numpy array.

        Returns:
            image (np.ndarray): the preprocessed image
            scale (float): the scale the image was resized to
        """
        return preprocess_image(image, self.image_size)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🎯 Human detection and tracking model that balances the importance between
detection and re-ID tasks.
"""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.model.fairmotv1 import fairmot_model
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Initializes and uses FairMOT tracking model to detect and track people
    from the supplied image frame.

    FairMOT is based on the anchor-free object detector CenterNet with
    modifications to balance the importance between detection and
    re-identification tasks in an object tracker.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_labels_data|

        |bbox_scores_data|

        |obj_attrs_data|
        :mod:`model.fairmot` produces the ``ids`` attribute which contains the
        tracking IDs of the detections.

    Configs:
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        score_threshold (:obj:`float`): **default = 0.5**. |br|
            Object confidence score threshold.
        K (:obj:`int`): **default = 500**. |br|
            Maximum number of objects output during the object detection stage.
        min_box_area (:obj:`int`): **default = 100**. |br|
            Minimum value for area of detected bounding box. Calculated by
            width * height.
        track_buffer (:obj:`int`): **default = 30**. |br|
            Threshold to remove track if track is lost for more frames
            than value.
        input_size (:obj:`List[int]`): **default = [864, 480]**. |br|
            Size (width, height) of the input image to the model. Raw
            video/image frames will be resized to the ``input_size`` before
            they are fed to the model.

    References:
        FairMOT: On the Fairness of Detection and Re-Identification in Multiple
        Object Tracking
        https://arxiv.org/abs/2004.01888

        Model weights trained by:
        https://github.com/ifzhang/FairMOT
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self._frame_rate = 30.0

        self.model = fairmot_model.FairMOTModel(self.config, self._frame_rate)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Tracks objects from image.

        Specifically for use with MOT evaluation, will attempt to get optional
        input `mot_metadata` and recreate `FairMOTModel` with the appropriate
        frame rate when necessary.

        Args:
            inputs (Dict[str, Any]): Dictionary with keys "img". When running
                under MOT evaluation, contains "mot_metadata" key as well.

        Returns:
            (Dict[str, Any]): Dictionary containing:
            - bboxes (List[np.ndarray]): Bounding boxes for tracked targets.
            - bbox_labels (np.ndarray): Bounding box labels, hard coded as
                "person".
            - bbox_scores (List[float]): Detection confidence scores.
            - obj_attrs (Dict[str, List[int]]): Tracking IDs, specifically for use
                with `mot_evaluator`.
        """
        metadata = inputs.get(
            "mot_metadata", {"frame_rate": self._frame_rate, "reset_model": False}
        )
        frame_rate = metadata["frame_rate"]
        reset_model = metadata["reset_model"]

        if frame_rate != self._frame_rate or reset_model:
            self._frame_rate = frame_rate
            self._reset_model()

        bboxes, bbox_scores, track_ids = self.model.predict(inputs["img"])
        bbox_labels = np.array(["person"] * len(bboxes))
        bboxes = np.clip(bboxes, 0, 1)

        outputs = {
            "bboxes": bboxes,
            "bbox_labels": bbox_labels,
            "bbox_scores": bbox_scores,
            "obj_attrs": {"ids": track_ids},
        }
        return outputs

    def _reset_model(self) -> None:
        """Creates a new instance of the FairMOT model with the frame rate
        supplied by `mot_metadata`.
        """
        self.logger.info(
            f"Creating new model with frame rate: {self._frame_rate:.2f}..."
        )
        self.model = fairmot_model.FairMOTModel(self.config, self._frame_rate)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Checks if detected objects are near each other.
"""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Checks if any objects are near each other.

    It does so by comparing the 3D locations of all objects to see which ones
    are near each other. If the distance between two objects is below the
    minimum threshold, both would be flagged as near with ``tag_msg``. These
    flags can be accessed by the ``flags`` key of :term:`obj_attrs`.

    Inputs:
        |obj_3D_locs_data|

    Outputs:
        |obj_attrs_data|
        :mod:`dabble.check_nearby_objs` produces the ``flags`` attribute which
        contains either the ``tag_msg`` for objects that are near each other or
        an empty string for objects with no other objects nearby.

    Configs:
        near_threshold (:obj:`float`): **default = 2.0**. |br|
            Threshold of distance, in metres, between two objects. Objects with
            distance less than ``near_threshold`` would be considered as 'near'.
        tag_msg (:obj:`str`): **default = "TOO CLOSE!"**. |br|
            Tag to identify objects which are near others.

    .. versionchanged:: 1.2.0
        :mod:`draw.check_nearby_objs` used to return ``obj_tags``
        (:obj:`List[str]`) as an output data type, which has been deprecated
        and now subsumed under :term:`obj_attrs`. The same attribute is
        accessed by using the ``flags`` key of :term:`obj_attrs`.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Compares the 3D locations of all objects to see which objects are
        close to each other.

        If an object is close to another, tag it.
        """
        obj_flags = [""] * len(inputs["obj_3D_locs"])

        for idx_1, loc_1 in enumerate(inputs["obj_3D_locs"]):
            for idx_2, loc_2 in enumerate(inputs["obj_3D_locs"]):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)
                if dist_bet < self.near_threshold:
                    obj_flags[idx_1] = self.tag_msg
                    break

        return {"obj_attrs": {"flags": obj_flags}}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Draws the bottom middle point of a bounding box."""

from typing import Any, Dict

from peekingduck.pipeline.nodes.draw.utils.bbox import draw_pts
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """The :mod:`draw.btm_midpoint` node uses :term:`bboxes` from the model
    predictions to draw the bottom midpoint of each bbox as a dot onto the
    image. For better understanding of the use case, refer to the
    :doc:`Zone Counting use case </use_cases/zone_counting>`.

    Inputs:
        |img_data|

        |btm_midpoint_data|

    Outputs:
        |none_output_data|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws btm midpoint of bounding bboxes.

        Args:
            inputs (dict): Dictionary with keys "bboxes".

        Returns:
            outputs (dict): Empty dictionary.
        """
        draw_pts(inputs["img"], inputs["btm_midpoint"])

        return {}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for the create node CLI command."""

import functools
import logging
import re
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import click
import yaml

# Master map file for class name to object IDs for object detection models
MASTER_MAP = "pipeline/nodes/model/master_map.yml"

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def create_config_and_script_files(created_paths: Dict[str, Path]) -> None:
    """Creates a config and script file at the provided paths. The contents of
    these files will be copied from the template config and script file.
    """
    created_paths["config"].parent.mkdir(parents=True, exist_ok=True)
    created_paths["script"].parent.mkdir(parents=True, exist_ok=True)
    template_paths = get_config_and_script_paths(
        Path(__file__).resolve().parents[1],
        "configs",
        ("pipeline", "nodes"),
        "node_template",
    )
    with open(template_paths["config"]) as template_file, open(
        created_paths["config"], "w"
    ) as outfile:
        outfile.write(template_file.read())
    with open(template_paths["script"]) as template_file, open(
        created_paths["script"], "w"
    ) as outfile:
        lines = template_file.readlines()
        start = -1
        for i, line in enumerate(lines):
            if line.startswith('"'):
                start = i
                break
        # In case we couldn't find a starting line
        start = max(0, start)
        outfile.writelines(lines[start:])


def ensure_relative_path(node_subdir: str) -> str:
    """Checks that the subdir path does not contain parent directory
    navigator (..), is not absolute, and is not "peekingduck/pipeline/nodes".
    """
    pkd_node_subdir = "peekingduck/pipeline/nodes"
    if ".." in node_subdir:
        raise click.exceptions.UsageError("Path cannot contain '..'!")
    if Path(node_subdir).is_absolute():
        raise click.exceptions.UsageError("Path cannot be absolute!")
    if node_subdir == pkd_node_subdir:
        raise click.exceptions.UsageError(f"Path cannot be '{pkd_node_subdir}'!")
    return node_subdir


def ensure_valid_name(node_dir: Path, node_type: str, node_name: str) -> str:
    """Checks the validity of the specified node_name. Also checks if it
    already exists.
    """
    if re.match(r"^[a-zA-Z][\w\-]*[^\W_]$", node_name) is None:
        raise click.exceptions.UsageError("Invalid node name!")
    if (node_dir / node_type / f"{node_name}.py").exists():
        raise click.exceptions.UsageError("Node name already exists!")
    return node_name


def ensure_valid_name_partial(node_dir: Path, node_type: str) -> Callable:
    """Partial function to ensure_valid_name to provide a function that matches
    function signature required by ``value_proc`` in ``click.prompt()``.
    """
    return functools.partial(ensure_valid_name, node_dir, node_type)


def ensure_valid_type(node_type_choices: click.Choice, node_type: str) -> str:
    """Uses click's convert_type function to check the validity of the
    specified node_type. Re-raises with a custom error message to ensure
    consistency across click versions.
    """
    try:
        click.types.convert_type(node_type_choices)(node_type)
    except click.BadParameter:
        raise click.exceptions.UsageError(
            f"'{node_type}' is not one of {', '.join(map(repr, node_type_choices.choices))}."
        ) from None
    return node_type


def ensure_valid_type_partial(node_type_choices: click.Choice) -> Callable:
    """Partial function to ensure_valid_type to provide a function that matches
    the function signature required by ``value_proc`` in ``click.prompt()``.
    """
    return functools.partial(ensure_valid_type, node_type_choices)


def get_config_and_script_paths(
    parent_dir: Path,
    config_subdir: Union[str, Tuple[str, ...]],
    script_subdir: Union[str, Tuple[str, ...]],
    file_stem: str,
) -> Dict[str, Path]:
    """Returns the node config file and its corresponding script file."""
    if isinstance(config_subdir, tuple):
        config_subpath = Path(*config_subdir)
    else:
        config_subpath = Path(config_subdir)
    if isinstance(script_subdir, tuple):
        script_subpath = Path(*script_subdir)
    else:
        script_subpath = Path(script_subdir)

    return {
        "config": parent_dir / config_subpath / f"{file_stem}.yml",
        "script": parent_dir / script_subpath / f"{file_stem}.py",
    }


def obj_det_load_class_id_mapping(node_name: str) -> Dict[str, int]:
    """Loads class name to object ID mapping from the file
    peekingduck/pipeline/nodes/model/master_map.yml
    for object detection models only.

    Tech Notes
        master_map.yml comprises two documents:
        1. model mapping: tells which mapping system a particular object detection model uses
        2. class name to ID mapping: maps class name to ID, supports multiple mapping systems

    Args:
        node_name (str): Tells function which mapping to load,
                            Possible values = { model.efficientdet, model.yolo, model.yolox }.

    Returns:
        Dict[str, int]: Mapping of class names to object IDs relevant to given node_name
    """
    # use __file__ instead of self._base_dir as latter can be set to any (temp) path
    # without `peekingduck/` subdirectory, resulting in master map file not found error
    master_map_file = Path(__file__).parents[1] / MASTER_MAP
    # read both documents from master_map.yml
    with master_map_file.open() as map_file:
        model_map_dict, class_id_map_dict = yaml.safe_load_all(map_file)

    # node_name sanity check, to preempt non-object detection model nodes
    node_name_list = list(map(lambda x: f"model.{x}", model_map_dict.keys()))
    assert (
        node_name in node_name_list
    ), f"Node Name Error: expect one of {node_name_list} but got {node_name}"

    # determine mapping system to use based on given model
    model = node_name.replace("model.", "")
    model_map = model_map_dict[model]

    # construct class name->object ID map for the model
    class_id_map = {}
    for key, val in class_id_map_dict.items():
        class_id = val[model_map]
        class_id_map[key.lower()] = class_id
    return class_id_map


def obj_det_change_class_name_to_id(
    node_name: str, key: str, value: List[Any]
) -> Tuple[str, List[int]]:
    """Process object detection model node's detect_ids key and check for
    any class names to be converted to object IDs.
    E.g. person to 0, car to 2

    Args:
        node_name (str): to determine which object detection model is being used
                         because different models can use different object IDs.
        key (str): expected to be "detect_ids"; error otherwise.
        value (List[Any]): list of class names or object IDs for detection.
                           If object IDs, do nothing.
                           If class names, convert to object IDs.

    Returns:
        Tuple[str, List[int]]: "detect_ids", list of sorted object IDs.
    """
    class_id_map = obj_det_load_class_id_mapping(node_name)

    if not value:
        logger.warning("detect_ids list is empty, defaulting to detect person")
        value = ["person"]
    elif value == ["*"]:
        logger.info("Detecting all object classes")
        value = [*class_id_map]

    value_lc = [x.lower() if isinstance(x, str) else x for x in value]

    # parse value_lc for possible class name errors
    invalid_class_names = []
    for class_name in value_lc:
        if isinstance(class_name, str) and class_name not in class_id_map:
            invalid_class_names.append(class_name)

    if invalid_class_names:
        logger.warning(f"Invalid class names: {invalid_class_names}")

    # convert class names to numeric object IDs, any errors default to zero
    obj_ids_set = {
        x if isinstance(x, int) else class_id_map.get(x, 0) for x in value_lc
    }
    obj_ids_sorted_list = sorted(list(obj_ids_set))
    return key, obj_ids_sorted_list


def verify_option(value: Optional[str], value_proc: Callable) -> Optional[str]:
    """Verifies that input value via click.option matches the expected value.

    This sets ``value`` to ``None`` if it is invalid so the rest of the prompt
    can flow smoothly.

    Args:
        value (Optional[str]): Input value.
        value_proc (Callable): A function to check the validity of ``value``.

    Returns:
        (Optional[str]): ``value`` if it is a valid value. ``None`` if it is
            not.

    Raises:
        click.exceptions.UsageError: When ``value`` is invalid.
    """
    if value is None:
        return value
    try:
        value = value_proc(value)
    except click.exceptions.UsageError as error:
        click.echo(f"Error: {error.message}", err=True)
        value = None
    return value
